from settings import *

API_URL = 'https://dogbex.ingresso.co.uk/cgi-bin/xml_core.exe'
EXT_START_SESSION_URL = 'https://dogbex.ingresso.co.uk/cgi-bin/xml_start_session.exe'

TEST_USERNAME = 'whitelabel-testuser'
TEST_PASSWORD = ''

TEST_CREDIT_USERNAME = 'credit-testuser'
TEST_CREDIT_PASSWORD = 'testing123'
